# -* coding: utf-8 -*
import torch
import time


def format_time(time_secs):
    return '{0:.3f}s'.format(time_secs)


def start_timer():
    """useful for benchmarking"""
    torch.cuda.synchronize()
    return time.perf_counter()


def summary(name, start_time, display=False):
    """useful for benchmarking"""
    torch.cuda.synchronize()
    end_time = time.perf_counter()
    step_time = end_time - start_time
    if display:
        print('{} time: {}'.format(name, format_time(step_time)))
    return step_time



import os
import subprocess
import matplotlib.pyplot as plt

IMGCAT = os.path.expanduser('~/local/bin/imgcat')


def zs_dispFig(imgcat_path=IMGCAT):
    """Shows a matplotlib plot inline in iTerm by saving it to a temporary file,
    displaying the file and then cleaning up.

    Note: This function requires the use of an iTerm terminal (available from
    https://www.iterm2.com) and requires that the `imgcat` script is on your $PATH. It
    can be useful when running code on a server when you don't have access to a GUI to
    display figures.

    Copyright (C) 2016 Samuel Albanie
    All rights reserved.
    """

    # save figure as temp image
    im_name = '_tmp.png'
    plt.savefig(im_name)

    # display in iterm
    subprocess.call([imgcat_path, "--depth", "iterm2", im_name])

    # clear up
    os.remove(im_name)



"""Datastructures for interacting with computer vision features.

The motivation for `ExpertStore` is to use memory backed by an array to avoid
fragmentation (rather than using a standard python dictionary), but still offer a
key-value accessor interface.

Relative comparison:
    Memory needs for dict: 417.1 MB
    Memory needs for np: 409.6 MB
    Memory needs for expert_store: 412.0 MB
"""

import argparse
import humanize
import pickle
import numpy as np
from typing import List


class ExpertStore:

    def __init__(self, keylist: List[str], dim: int, dtype: np.dtype = np.float16):
        self.keys = keylist
        self.dim = dim
        self.store_dtype = dtype
        self.store = np.zeros((len(keylist), dim), dtype=dtype)
        self.keymap = {}
        self.missing = set()

        for idx, key in enumerate(keylist):
            self.keymap[key] = idx

    def __setitem__(self, key, value):
        idx = self.keymap[key]
        if isinstance(value, np.ndarray):
            # non-nan values must be vectors of the appropriate size
            assert value.size == self.dim, f"cannot set value with size {value.size}"
        else:
            assert np.isnan(value)
        self.store[idx] = value

    def __getitem__(self, key):
        return self.store[self.keymap[key]]

    def todict(self):
        """Convert the current datastructure into a vanilla python dictionary

        Returns:
            a dictionary with the same keys and values as the current object.
        """
        return {key: self[key] for key in self.keymap}

    def __len__(self):
        return len(self.keys)

    def __repr__(self):
        keep_samples = 3
        samples = list(self.keymap.items())[:keep_samples]
        sample_str = "\n".join([f"{key}: {val}" for key, val in samples])
        summary = (
            f"ExpertStore object with {len(self.keys)} features (dim: {self.dim})"
            f" (storage is using {humanize.naturalsize(self.store.nbytes)})"
            f"\nFirst {keep_samples} elements of keymap: \n{sample_str}"
        )
        return summary


def gen_dict_store(keylist, dim):
    store = dict()
    for key in keylist:
        store[key] = np.random.rand(1, dim).astype(np.float16)
    return store


class HashableDict(dict):
    def __hash__(self):
        return hash(frozenset(self))


class HashableOrderedDict(dict):
    def __hash__(self):
        return hash(frozenset(self))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--dim", type=int, default=2048)
    parser.add_argument("--num_keys", type=int, default=100000)
    args = parser.parse_args()
    keylist = [str(x) for x in range(args.num_keys)]
    for store_name in "dict", "np", "expert_store":
        if store_name == "dict":
            store = gen_dict_store(keylist=keylist, dim=args.dim)
        elif store_name == "np":
            store = np.random.rand(len(keylist), args.dim).astype(np.float16)
        elif store_name == "expert_store":
            store = ExpertStore(keylist=keylist, dim=args.dim)
            print(store)
        serialised = pickle.dumps(store)
        print(f"Memory used by {store_name}: {humanize.naturalsize(len(serialised))}")


if __name__ == "__main__":
    main()



import sys
import time
import pickle
import socket
import json
import functools
from typing import Union
from pathlib import Path
import scipy.io as spio
from typeguard import typechecked

import numpy as np
import msgpack_numpy as msgpack_np
import zsvision.zs_data_structures

msgpack_np.patch()


@functools.lru_cache(maxsize=64, typed=False)
@typechecked
def memcache(path: Union[Path, str]):
    path = Path(path)
    suffix = path.suffix
    print(f"loading data from {path} ({socket.gethostname()})", end=" ", flush=True)
    tic = time.time()
    if suffix in {".pkl", ".pickle"}:
        res = pickle_loader(path)
    elif suffix == ".npy":
        res = np_loader(path)
    elif suffix == ".mp":
        res = msgpack_loader(path)
    elif suffix == ".json":
        with open(path, "r") as f:
            res = json.load(f)
    elif suffix == ".mat":
        res = loadmat(path)
    else:
        raise ValueError(f"unknown suffix: {suffix} for path {path}")
    print(f"[Total: {time.time() - tic:.1f}s]")
    return res


@typechecked
def support_old_pickles(buffer: bytes) -> object:
    try:
        data = pickle.loads(buffer, encoding="latin1")
    except ModuleNotFoundError as exception:
        if "datastructures" in exception.msg:
            sys.modules['datastructures'] = zsvision.zs_data_structures
            data = pickle.loads(buffer, encoding="latin1")
    return data


@typechecked
def pickle_loader(pkl_path: Path, backwards_compatible: bool = True) -> object:
    """Deserialise object from pickle.

    Args:
        pkl_path: the location of the path where the pickle path is stored
        backwards_compatible: if true, support old pickle formats used with the.
            ExpertStore format

    Return:
        The deserialised object.
    """
    tic = time.time()
    with open(pkl_path, "rb") as f:
        buffer = f.read()
        print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
        tic = time.time()
        if backwards_compatible:
            data = support_old_pickles(buffer)
        else:
            data = pickle.loads(buffer, encoding="latin1")
        print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
    return data


@typechecked
def msgpack_loader(mp_path: Path):
    """Msgpack provides a faster serialisation routine than pickle, so is preferable
    for loading and deserialising large feature sets from disk."""
    tic = time.time()
    with open(mp_path, "rb") as f:
        buffer = f.read()
        print(f"[I/O: {time.time() - tic:.1f}s]", end=" ")
        tic = time.time()
        data = msgpack_np.unpackb(buffer, object_hook=msgpack_np.decode, encoding="utf-8")
        print(f"[deserialisation: {time.time() - tic:.1f}s]", end=" ")
    return data


def np_loader(np_path, l2norm=False):
    with open(np_path, "rb") as f:
        data = np.load(f, encoding="latin1", allow_pickle=True)
    if isinstance(data, np.ndarray) and data.size == 1:
        data = data[()]  # handle numpy dict storage convnetion
    if l2norm:
        print("L2 normalizing features")
        if isinstance(data, dict):
            for key in data:
                feats_ = data[key]
                feats_ = feats_ / max(np.linalg.norm(feats_), 1E-6)
                data[key] = feats_
        elif data.ndim == 2:
            data_norm = np.linalg.norm(data, axis=1)
            data = data / np.maximum(data_norm.reshape(-1, 1), 1E-6)
        else:
            raise ValueError("unexpected data format {}".format(type(data)))
    return data


def set_nested_key_val(key, val, target):
    """Use a prefix key (e.g. key1.key2.key3) to set a value in a nested dict"""

    # escape periods in keys
    key = key.replace("_.", "&&")
    subkeys = key.split(".")
    subkeys = [x.replace("&&", ".") for x in subkeys]

    nested = target
    print("subkeys", subkeys)
    for subkey in subkeys[:-1]:
        try:
            nested = nested.__getitem__(subkey)
        except Exception as exception:
            print(subkey)
            raise exception
    orig = nested[subkeys[-1]]
    if orig == "":
        if val == "":
            val = 0
        else:
            val = str(val)
    elif isinstance(orig, bool):
        if val.lower() in {"0", "False"}:
            val = False
        else:
            val = bool(val)
    elif isinstance(orig, list):
        if isinstance(val, str) and "," in val:
            val = val.split(",")
            # we use the convention that a trailing comma indicates a single item list
            if len(val) == 2 and val[1] == "":
                val.pop()
            if val and not orig:
                raise ValueError(f"Could not infer correct type from empty original list")
            else:
                val = [type(orig[0])(x) for x in val]
        assert isinstance(val, list), "Failed to pass a list where expected"
    elif isinstance(orig, int):
        val = int(val)
    elif isinstance(orig, float):
        val = float(val)
    elif isinstance(orig, str):
        val = str(val)
    else:
        raise ValueError(f"unrecognised type: {type(val)}")
    nested[subkeys[-1]] = val


def loadmat(filename):
    """This function should be called instead of direct spio.loadmat
    as it cures the problem of not properly recovering python dictionaries
    from mat files. It calls the function check keys to cure all entries
    which are still mat-objects.

    The function is heavily based on this reference:
    https://stackoverflow.com/a/29126361

    Args:
        filename (str): the location of the .mat file to load
    
    Returns:
        (dict): a parsed .mat file in the form of a python dictionary.
    """
    def _check_keys(d):
        """Checks if entries in dictionary are mat-objects. If yes
        todict is called to change them to nested dictionaries
        """
        for key in d:
            if isinstance(d[key], spio.matlab.mio5_params.mat_struct):
                d[key] = _todict(d[key])
            elif isinstance(d[key], np.ndarray):
                d[key] = _tolist(d[key])
            else:
                pass
        return d

    def _todict(matobj):
        """A recursive function which constructs from matobjects nested dictionaries
        """
        d = {}
        for strg in matobj._fieldnames:
            elem = matobj.__dict__[strg]
            if isinstance(elem, spio.matlab.mio5_params.mat_struct):
                d[strg] = _todict(elem)
            elif isinstance(elem, np.ndarray):
                d[strg] = _tolist(elem)
            else:
                d[strg] = elem
        return d

    def _tolist(ndarray):
        """A recursive function which constructs lists from cellarrays
        (which are loaded as numpy ndarrays), recursing into the elements
        if they contain matobjects or are non-numeric.
        """
        if np.issubdtype(ndarray.dtype, np.number):
            return ndarray
        elem_list = []
        for sub_elem in ndarray:
            if isinstance(sub_elem, spio.matlab.mio5_params.mat_struct):
                elem_list.append(_todict(sub_elem))
            elif isinstance(sub_elem, np.ndarray):
                elem_list.append(_tolist(sub_elem))
            else:
                elem_list.append(sub_elem)
        return elem_list

    data = spio.loadmat(filename, struct_as_record=False, squeeze_me=True)
    return _check_keys(data)


@functools.lru_cache(maxsize=64, typed=False)
def concat_features(feat_paths, axis):
    aggregates = [memcache(x) for x in feat_paths]
    tic = time.time()
    msg = "expected to concatenate datastructures of a single type"
    assert len(set(type(x) for x in aggregates)) == 1, msg
    if isinstance(aggregates[0], dict):
        keys = aggregates[0]  # for now, we assume that all aggregates share keys
        merged = {}
        for key in keys:
            merged[key] = np.concatenate([x[key] for x in aggregates], axis=axis)
    elif isinstance(aggregates[0], zsvision.zs_data_structures.ExpertStore):
        dims, stores = [], []
        keys = aggregates[0].keys
        for x in aggregates:
            dims.append(x.dim)
            stores.append(x.store)
            try:
                assert x.keys == keys, "all aggregates must share identical keys"
            except Exception as E:
                print(E)
                import ipdb; ipdb.set_trace()
        msg = "expected to concatenate ExpertStores with a common dimension"
        assert len(set(dims)) == 1, msg
        dim = dims[0]
        merged = zsvision.zs_data_structures.ExpertStore(keys, dim=dim)
        merged.store = np.concatenate(stores, axis=axis)
    else:
        raise ValueError(f"Unknown datastructure: {type(aggregates[0])}")
    # Force memory clearance
    for aggregate in aggregates:
        del aggregate
    print("done in {:.3f}s".format(time.time() - tic))
    return merged


class BlockTimer:
    """A minimal inline codeblock timer"""
    def __init__(self, msg, precise=False, mute=False):
        self.msg = msg
        self.mute = mute
        self.precise = precise

    def __enter__(self):
        self.start = time.time()
        print(f"{self.msg}...", end="", flush=True)
        return self

    def __exit__(self, *args):
        if self.precise:
            total = f"{time.time() - self.start:.3f}s"
        else:
            total = time.strftime('%Hh%Mm%Ss', time.gmtime(time.time() - self.start))
        if not self.mute:
            print(f" took {total}")



"""Utilities for multiprocessing with Python.

NOTE: The starmap kwargs mapper is based on isarandi's StackOverflow answer here:
stackoverflow.com/questions/45718523/pass-kwargs-to-starmap-while-using-pool-in-python
"""
from itertools import repeat


def apply_kwargs(func, kwargs):
    """Wrapper for unpacking keyword function calls.

    Args:
        func (function): the function to be applied.
        kwargs (dict): the keywords and arguments for the function call.
    """
    return func(**kwargs)


def starmap_with_kwargs(pool, func, kwargs_iter):
    """Apply starmap with keyword arguments.

    Args:
        pool (multiprocessing.Pool): A pool of processes.
        func (function): the function to be applied.
        kwargs_iter (list[dict]): a list of dictionaries, each of which contains the
            keywords for a separate function call.
    """
    args_for_starmap = zip(repeat(func), kwargs_iter)
    return pool.starmap(apply_kwargs, args_for_starmap)
